const {Command, flags} = require('@oclif/command')
const {cli} = require('cli-ux')
const axios = require('axios')
const chalk = require('chalk')
const inquirer = require('inquirer')
const prompt = inquirer.createPromptModule()
const execa = require('execa')
const listr = require('listr')
const fs = require('fs-extra')
const path = require('path')
const mysql = require('mysql')



class PagesPayrollCliCommand extends Command {

  async run() {

    // create config directory if it does not exist
    if (!fs.existsSync(this.config.configDir)){
      fs.mkdirSync(this.config.configDir);
    }
    // read the config from configDir
    const userConfig = await fs.readJSON(path.join(this.config.configDir, 'config.json'))
    
    const {flags, args} = this.parse(PagesPayrollCliCommand)
    const do_headers = {
      headers: {
        'Content-Type': 'application/json',
        'Authorization': `Bearer ${userConfig.DO_ACCESS_TOKEN}`
      }
    }
    const sentry_headers = {
      headers: {
        'Content-Type': 'application/json',
        'Authorization': `Bearer ${userConfig.SENTRY_TOKEN}`
      }
    }

    let hostname = flags.hostname || args.hostname || null
    let sentry_dsn = null
    
    const _self = this
    
    // if(!hostname || !hostname.match(/^[a-z]{3,15}$/i)){
      const response = await prompt([
        // get hostname
        {
          type: 'input',
          name: 'hostname',
          message: chalk`Enter the {bold.yellow hostname}: `,
          default: hostname,
          validate: function(value) {
            return value.match(/^[a-z]{3,15}$/i) ? true : 'Please enter a valid hostname!';
          },
          choices: [],
        },
        // get app_name
        {
          type: 'input',
          name: 'app_name',
          message: chalk`Enter the {bold.yellow APP_NAME}: `,
          default: `Payroll ${hostname}`,
          choices: [],
        },
        // get app_url
        {
          type: 'input',
          name: 'app_url',
          message: chalk`Enter the {bold.yellow APP_URL}: `,
          default: `https://${hostname}.${userConfig.DOMAIN_NAME}`,
          choices: [],
        },
        // get db_host
        {
          type: 'input',
          name: 'db_host',
          message: chalk`Enter the {bold.yellow DB_HOST}: `,
          default: userConfig.DB_HOST,
          choices: [],
        },
        // get db_port
        {
          type: 'input',
          name: 'db_port',
          message: chalk`Enter the {bold.yellow DB_PORT}: `,
          default: userConfig.DB_PORT,
          choices: [],
        },
        // get db_username
        {
          type: 'input',
          name: 'db_username',
          message: chalk`Enter the {bold.yellow DB_USERNAME}: `,
          default: userConfig.DB_USERNAME,
          choices: [],
        },
        // get db_password
        {
          type: 'input',
          name: 'db_password',
          message: chalk`Enter the {bold.yellow DB_PASSWORD}: `,
          default: userConfig.DB_PASSWORD,
          choices: [],
        },
        // get mail_username
        {
          type: 'input',
          name: 'mail_username',
          message: chalk`Enter the {bold.yellow MAIL_USERNAME}: `,
          default: userConfig.MAIL_USERNAME,
          choices: [],
        },
        // get mail_password
        {
          type: 'input',
          name: 'mail_password',
          message: chalk`Enter the {bold.yellow MAIL_PASSWORD}: `,
          default: userConfig.MAIL_PASSWORD,
          choices: [],
        },
        // get mail_from_name
        {
          type: 'input',
          name: 'mail_from_name',
          message: chalk`Enter the {bold.yellow MAIL_FROM_NAME}: `,
          default: 'Payroll Admin',
          choices: [],
        },
        // get mail_from_address
        {
          type: 'input',
          name: 'mail_from_address',
          message: chalk`Enter the {bold.yellow MAIL_FROM_ADDRESS}: `,
          default: `info@${userConfig.DOMAIN_NAME}`,
          choices: [],
        },
        // get msg_key
        {
          type: 'input',
          name: 'msg_key',
          message: chalk`Enter the {bold.yellow MSG_KEY}: `,
          choices: [],
        },
        // get msg_sender_id
        {
          type: 'input',
          name: 'msg_sender_id',
          message: chalk`Enter the {bold.yellow MSG_SENDER_ID}: `,
          choices: [],
        },
        // get key_id
        {
          type: 'input',
          name: 'key_id',
          message: chalk`Enter the {bold.yellow AWS KEY ID}: `,
          default: `${userConfig.AWS_KEY_ID}`,
          choices: [],
        },
        // get access_key
        {
          type: 'input',
          name: 'access_key',
          message: chalk`Enter the {bold.yellow AWS ACCESS KEY}: `,
          default: `${userConfig.AWS_ACCESS_KEY}`,
          choices: [],
        }
      ])
      hostname = response.hostname
    // }
    
    // Steps for Pages Payroll Deployment
    const tasks = new listr(
      [
        // Create a new A record for new PagesPayroll Client
        {
          title: `Create a new subdomain - ${hostname}.${userConfig.DOMAIN_NAME}`,
          task: () => {
            return new listr([
              // check if subdomain exists
              {
                title: 'Checking if subdomain already exists',
                task: async (ctx, task) => {
                  await axios.get(`https://api.digitalocean.com/v2/domains/${userConfig.DOMAIN_NAME}/records`, do_headers)
                    .then(function (response) {
                        // handle success
                        if(response.data.domain_records.some((record) => record.name === hostname)){
                          ctx.hostname_exists = true
                        }
                        else{
                          ctx.hostname_exists = false
                        }
                    })
                    .catch(function (error) {
                        // handle error
                        throw new Error(error.response.data.message)
                    })
                }
              },
              // create a new subdomain
              {
                title: `Create new subdomain - ${hostname}.${userConfig.DOMAIN_NAME}`,
                skip: (ctx) => {
                  return ctx.hostname_exists ? `Subdomain - ${hostname}.${userConfig.DOMAIN_NAME} already exists!` : false
                },
                task: (ctx, task) => {
                  const data = {
                    type: 'A',
                    name: hostname,
                    data: userConfig.DOMAIN_IP
                  }
                  axios.post(`https://api.digitalocean.com/v2/domains/${userConfig.DOMAIN_NAME}/records`, data, do_headers)
                    .then(function (response) {
                        // handle success
                        return `${hostname}.${userConfig.DOMAIN_NAME} created with ID: ${response.data.id}`
                    })
                    .catch(function (error) {
                        // handle error
                        throw new Error(error.response.data.message)
                    })
                }
              }
            ])
          }
        },
        // Create a new Sentry Project
        {
          title: `Create a new sentry project - ${hostname}-payroll and Get Client Keys`,
          task: () => {
            return new listr([
              // check if the project already exits in sentry
              {
                title: `Check if ${hostname}-payroll exists in Sentry`,
                task: async (ctx, task) => {
                  await axios.get(`https://sentry.io/api/0/teams/three38-innovation-cafe-llp/pages-payroll/projects/`, sentry_headers)
                    .then(function (response) {
                        // handle success
                        if(response.data.some((record) => record.slug == `${hostname}-payroll`)){
                          ctx.project_exists = true
                        }
                        else{
                          ctx.project_exists = false
                        }
                    })
                    .catch(function (error) {
                        // handle error
                        throw new Error(error.response.data.message)
                    })
                }
              },
              // create a new sentry project
              {
                title: `Create new sentry project - ${hostname}-payroll`,
                skip: (ctx) => {
                  return ctx.project_exists ? `Sentry project - ${hostname}-payroll already exists!` : false
                },
                task: async (ctx, task) => {
                  const data = {
                      name: `${hostname} Payroll`,
                      slug: `${hostname}-payroll`,
                      platform: "php-laravel"
                  }
                  await axios.post(`https://sentry.io/api/0/teams/three38-innovation-cafe-llp/pages-payroll/projects/`, data, sentry_headers)
                              .catch(function (error) {
                                  // handle error
                                  throw new Error(error.response.data.message)
                              })
                }
              },
              // get sentry project client keys
              {
                title: `Get client keys for ${hostname}-payroll`,
                skip: (ctx) => {
                  return ctx.project_exists ? `Sentry project - ${hostname}-payroll already exists!` : false
                },
                task: async (ctx, task) => {
                  await axios.get(`https://sentry.io/api/0/projects/three38-innovation-cafe-llp/${hostname}-payroll/keys/`, sentry_headers)
                    .then(function (response) {
                        // handle success
                        // get the key with name as Default and then get the DSN public key
                        let default_key = response.data.find(key => key.name == 'Default')
                        if(default_key){
                          sentry_dsn = default_key.dsn.public
                        }
                        else{
                          throw new Error('Default key not found.')
                        }
                        return `${hostname}-payroll default key: ${response.data.id}`
                    })
                    .catch(function (error) {
                        // handle error
                        throw new Error(error.response.data.message)
                    })
                }
              }
            ])
          }
        },
        // Create virtual host conf in apache2 sites-available 
        {
          title: `Create virtual host configurations - ${hostname}.${userConfig.DOMAIN_NAME}.conf`,
          task: () => {
            return new listr([
              // create a copy of the conf file
              {
                title: `Create a copy of the copy.${userConfig.DOMAIN_NAME}.conf`,
                task: async (ctx, task) => {
                  await execa('cp', [`${userConfig.APACHE_DIR}/copy.${userConfig.DOMAIN_NAME}.conf`, `${userConfig.APACHE_DIR}/${hostname}.${userConfig.DOMAIN_NAME}.conf`])
                  return 'success'
                }
              },
              // replace demo.conf with subdomain.conf
              {
                title: `Replace 'demo.${userConfig.DOMAIN_NAME}' with '${hostname}.${userConfig.DOMAIN_NAME}' in the conf file`,
                task: (ctx, task) => {
                  let filename = `${userConfig.APACHE_DIR}/${hostname}.${userConfig.DOMAIN_NAME}.conf`
                  fs.readFile(filename, 'utf8', function (err, data) {
                    if (err) {
                      return console.log(err);
                    }
                    var result = data.replace(/demo/g, hostname)

                    fs.writeFile(filename, result, 'utf8', function (err) {
                      if (err) return console.log(err);
                    })
                  })
                  return `conf file updated`
                }
              },
              // enable the new subdomain.conf
              {
                title: `Enable the ${hostname}.${userConfig.DOMAIN_NAME}.conf`,
                task: async (ctx, task) => {
                  process.chdir(`${userConfig.APACHE_DIR}`)
                  await execa('a2ensite', [`${hostname}.${userConfig.DOMAIN_NAME}.conf`])
                  return 'success'
                }
              },
              // restart apache2
              {
                title: `Restart apache2`,
                task: async (ctx, task) => {
                  await execa('systemctl', ['restart', 'apache2'])
                  return 'success'
                }
              },
              // restart apache2 service
              {
                title: `Restart apache2 service`,
                task: async (ctx, task) => {
                  await execa('service', ['apache2', 'restart'])
                  return 'success'
                }
              }
            ])
          }
        },
        // Create a new subdirectory and update the current working directory & Clone Code from BitBucket
        {
          title: `Create a new subdirectory & Clone Code from BitBucket`,
          task: () => {
            return new listr([
              // create a new sub directory
              {
                title: `Create a new subdirectory - ${userConfig.HOME_DIR}/${hostname}.${userConfig.DOMAIN_NAME}`,
                task: (ctx, task) => {
                  var dir = `${userConfig.HOME_DIR}/${hostname}.${userConfig.DOMAIN_NAME}`;
                  if (!fs.existsSync(dir)){
                    fs.mkdirSync(dir);
                    return 'success'
                  }
                }
              },
              // change the current working directory
              {
                title: `Change the current working directory`,
                task: (ctx, task) => {
                  // change directory to the newly created subdirectory
                  process.chdir(`${userConfig.HOME_DIR}/${hostname}.${userConfig.DOMAIN_NAME}`);
                  return `Changed directory to: ${process.cwd()}`
                }
              },
              // clone the bitbucket repo
              {
                title: `Clone the BitBucket Repository`,
                task: async (ctx, task) => {
                  await execa('git', ['clone', userConfig.REPOSITORY_URL, '.'])
                  return 'success'
                }
              }
            ])
          }
        },
        // Setup a new MySQL Database for the client
        {
          title: `Setup new MySQL Database - pages_${hostname}`,
          task: (ctx, task) => {
            let conn = mysql.createConnection({
              host: response.db_host,
              port: response.db_port,
              user: response.db_username,
              password: response.db_password
            })
            // create a new database
            conn.connect(function(err) {
              if (err) throw err;
              conn.query(`CREATE DATABASE pages_${hostname}`, function (err, result) {
                if (err) throw err;
                return 'success'
              })
            }) 
          }
        },
        // Setup Laravel
        {
          title: `Setup Laravel`,
          task: () => {
            return new listr([
              // change the current working directory
              {
                title: `Change the current working directory`,
                task: (ctx, task) => {
                  // change directory to the newly created subdirectory
                  process.chdir(`${userConfig.HOME_DIR}/${hostname}.${userConfig.DOMAIN_NAME}`);
                  return `Changed directory to: ${process.cwd()}`
                }
              },
              // create the .env file from .env.example
              {
                title: `Create the .env file from .env.example`,
                task: async (ctx, task) => {
                  await execa('cp', ['.env.example', '.env'])
                  return 'success'
                }
              },
              // make the changes in the .env file
              {
                title: `Make changes in the .env file`,
                task: (ctx, task) => {
                 
                  let filename = `.env`
                  
                  fs.readFile(filename, 'utf8', function (err, data) {
                    if (err) {
                      return console.log(err);
                    }
                    let result = data.replace(/DEMO_APPNAME/g, response.app_name)
                    result = result.replace(/DEMO_NAME/g, hostname)
                    result = result.replace(/DEMO_URL/g, response.app_url)
                    result = result.replace(/DEMO_HOST/g, response.db_host)
                    result = result.replace(/DEMO_PORT/g, response.db_port)
                    result = result.replace(/DEMO_DATABASE/g, `pages_${hostname}`)
                    result = result.replace(/DEMO_USERNAME/g, response.db_username)
                    result = result.replace(/DEMO_PASSWORD/g, response.db_password)
                    result = result.replace(/DEMO_MAIL_USERNAME/g, response.mail_username)
                    result = result.replace(/DEMO_MAIL_PASSWORD/g, response.mail_password)
                    result = result.replace(/DEMO_FROM_NAME/g, response.mail_from_name)
                    result = result.replace(/DEMO_FROM_ADDRESS/g, response.mail_from_address)
                    result = result.replace(/DEMO_MSG_KEY/g, response.msg_key)
                    result = result.replace(/DEMO_MSG_SENDER_ID/g, response.msg_sender_id)
                    result = result.replace(/DEMO_KEY_ID/g, response.key_id)
                    result = result.replace(/DEMO_ACCESS_KEY/g, response.access_key)
                    result = result.replace(/DEMO_LARAVEL_DSN/g, sentry_dsn)
                    result = result.replace(/DEMO_MIX_DSN/g, sentry_dsn)

                    fs.writeFile(filename, result, 'utf8', function (err) {
                      if (err) return console.log(err);
                    })
                  })
                  return `env file updated`
                }
              },
              // composer install
              {
                title: `Composer install`,
                task: async (ctx, task) => {
                  await execa('composer', ['install'])
                  return 'success'
                }
              },
              // settings permissions
              {
                title: `Setting Permissions`,
                task: () => {
                  return new listr([
                    {
                      title: `chown storage`,
                      task: async (ctx, task) => {
                        await execa('chown', ['-R', 'root:www-data', 'storage'])
                        return 'success'
                      }
                    },
                    {
                      title: `chown bootstrap/cache`,
                      task: async (ctx, task) => {
                        await execa('chown', ['-R', 'root:www-data', 'bootstrap/cache'])
                        return 'success'
                      }
                    },
                    {
                      title: `chmod storage`,
                      task: async (ctx, task) => {
                        await execa('chmod', ['-R', '775', 'storage'])
                        return 'success'
                      }
                    },
                    {
                      title: `chmod bootstrap/cache`,
                      task: async (ctx, task) => {
                        await execa('chmod', ['-R', '775', 'bootstrap/cache'])
                        return 'success'
                      }
                    },
                    {
                      title: `chown public/uploads`,
                      task: async (ctx, task) => {
                        await execa('chmod', ['-R', '775', 'public/uploads'])
                        return 'success'
                      }
                    },
                  ])
                }
              },
              // run artisan commands
              {
                title: `Run Artisan Commands`,
                task: () => {
                  return new listr([
                    {
                      title: `php artisan migrate`,
                      task: async (ctx, task) => {
                        await execa('php', ['artisan', 'migrate'])
                        return 'success'
                      }
                    },
                    {
                      title: `php artisan db:seed`,
                      task: async (ctx, task) => {
                        await execa('php', ['artisan', 'db:seed'])
                        return 'success'
                      }
                    },
                    {
                      title: `php artisan passport:install`,
                      task: async (ctx, task) => {
                        await execa('php', ['artisan', 'passport:install'])
                        return 'success'
                      }
                    },
                    {
                      title: `php artisan route:cache`,
                      task: async (ctx, task) => {
                        await execa('php', ['artisan', 'route:cache'])
                        return 'success'
                      }
                    },
                    {
                      title: `php artisan config:cache`,
                      task: async (ctx, task) => {
                        await execa('php', ['artisan', 'config:cache'])
                        return 'success'
                      }
                    },
                    {
                      title: `php artisan view:cache`,
                      task: async (ctx, task) => {
                        await execa('php', ['artisan', 'view:cache'])
                        return 'success'
                      }
                    },
                  ])
                }
              }
            ])
          }
        },
        // setting up Https with Certbot
        // composer install
        {
          title: `Genrate HTTPS Certificate`,
          task: async (ctx, task) => {
            await execa('certbot --apache -n', [`-d ${hostname}.${userConfig.DOMAIN_NAME}`])
            return 'success'
          }
        },
      ]
    )

    // Run the Tasks
    tasks.run()
      .then(result => {
        _self.log(chalk`{bold.green Deployment Complete!}`)
        process.exit()
      })
      .catch(err => {
        this.error(err)
      })
  }

}

PagesPayrollCliCommand.description = `Automatic Deployment of Pages Payroll Web Application
...

`

PagesPayrollCliCommand.flags = {
  // add --version flag to show CLI version
  version: flags.version({char: 'v'}),
  // add --help flag to show CLI version
  help: flags.help({char: 'h'}),
  hostname: flags.string({char: 'hn'})
}

PagesPayrollCliCommand.args = [
  {
    name: 'hostname',             // name of arg to show in help and reference with args[name]
    required: false,              // make the arg required with `required: true`
    description: 'hostname for pagespayroll.app subdomain', // help description
    hidden: false,                // hide this arg from help
  }
]

module.exports = PagesPayrollCliCommand